import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import processing.video.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class DelayVideo extends PApplet {

/*
A mix of delayed images to the screen
 blend-delay via a map!
 */


Capture camera;
static int fps=15;
/* save three seconds */

PImage timeBuffer[];
final int timeBufferSize=60;
final int likelySize=4;

// was 16
// doesn't really matter
final int blendLevels=16;
final int blendLevelsL2=4;
final int blendLevelsMASK=0x0000000F;

// roughly scale a 4x3 image
// multiples of 20 only?
final int scale = 120;
final int screenscale = 1;

final int vvw = 4*scale; // video width
final int vvh = 3*scale; // video height

final int svw = 4*screenscale*scale; // screen width
final int svh = 3*screenscale*scale; // screen height

final int vwo2=vvw/2;
final int vho2=vvh/2;
int cw,ch;

boolean newFrame = false;
/*
Processing 2.1 version
*/
int saving=0;
int iBuffer;
int blendMap[];

int functionType;
int blendType;
boolean checkery;
PFont myFont; 


public void setup()
{
  int b;
  // 16 for label at the bottom
  size(svw, svh+16);
String[] cameras = Capture.list();

 if (cameras.length == 0) 
 {
    println("There are no cameras available for capture.");
    exit();
  }
   println("Available cameras:");
    for (int i = 0; i < cameras.length; i++) {
      println(cameras[i]);
    }
    
  camera = new Capture(this, vvw, vvh,"Built-in iSight", fps);
//camera = new Capture(this,cameras[1]);
 camera.start();

  timeBuffer= new PImage[timeBufferSize];
  // initialize time images
  for(int i=0;i<timeBufferSize;i++)
    timeBuffer[i] = new PImage(vvw,vvh);
    
  iBuffer=0;
  b=0;  
  cw=vwo2;
  ch=vho2;
  // String[] fontList = PFont.list(); 
  //println(fontList); 
  myFont = createFont("Arial", 14); 
  fill(64,64,64);
  textFont(myFont);
  functionType=0;
  blendType=0;
  checkery=false;
  setupBlendMap();

}

// put scaled fps/pixel in this array
public void setupBlendMap()
{
  float centerTime=0.0f;
  float edgeTime=1.0f;
  float r;
  float dx,dy;
  int ac,dn,coord;
  int tbuf,tbuf0;
  int likelyIndex;
  int likelyColorB;
  int likelyVariance;
 int thatBrt;
 
  blendMap= new int[vvw*vvh];

  // tubular map.
  if(functionType ==0)
  {
    for( dn=0;dn<vvh;dn++)
      for( ac=0;ac<vvw;ac++)
      {
        coord=dn*vvw+ac;
        // some kind of function for fps
        // center should be slow, edge should be fast!
        dx=(float)(ac-cw)/(float)vvw;
        dy=(float)(dn-ch)/(float)vvh;
        r=sqrt(dx*dx+dy*dy);
        // blend amounts... hmmm!
        // the center should blend the most.. about 5%new to 95% old, edge should be 100% new.
        // center == 0, edge can be anything over 1
        if(r<0.1f)
        {
          r=centerTime;
        }
        else
        {
          r-=0.1f;
          r=pow(lerp(centerTime,edgeTime,min(r,0.99f)),0.6f);
        } // vw debug
        blendMap[coord]=(int) (((timeBufferSize-1)*blendLevels)*min(1.0f,r)) ;


        //if(r<0.01)
        //  println("r:"+r+" dx:"+dx+" dy:"+dy+" bm:"+blendMap[coord]);
      }

  }
  // left - right map
  if(functionType ==1)
  {
    for( dn=0;dn<vvh;dn++)
      for( ac=0;ac<vvw;ac++)
      {
        coord=dn*vvw+ac;
        // some kind of function for fps
        // center should be slow, edge should be fast!
        dx=(float)(ac-vwo2)/(float)vwo2;

        r=abs(dx) ;
        // blend amounts... hmmm!
        // the center should blend the most.. about 5%new to 95% old, edge should be 100% new.
        // center == 0, edge can be anything over 1

        r= lerp(centerTime,edgeTime,min(r,0.99f)) ;

        blendMap[coord]=(int) (((timeBufferSize-1)*blendLevels)*min(1.0f,r)) ;

        //if(r<0.01)
        //  println("r:"+r+" dx:"+dx+" dy:"+dy+" bm:"+blendMap[coord]);
      }


  }

  // top - bottom map
  if(functionType ==2)
  {
    for( dn=0;dn<vvh;dn++)
      for( ac=0;ac<vvw;ac++)
      {
        coord=dn*vvw+ac;
        // some kind of function for fps
        // center should be slow, edge should be fast!


        r=(float)dn/(float)vvh;
        // blend amounts... hmmm!
        // the center should blend the most.. about 5%new to 95% old, edge should be 100% new.
        // center == 0, edge can be anything over 1

        r= lerp(centerTime,edgeTime,max(0.0f,min(r,0.99f))) ;

        blendMap[coord]=(int) (((timeBufferSize-1)*blendLevels)*min(1.0f,r)) ;
        //  if(r<0.10)
        //  println("r:"+r+" map:"+blendMap[coord]);
      }


  }

  // image brightness map
  if(functionType ==3)
  {
    for( dn=0;dn<vvh*vvw;dn+=vvw)
      for( ac=0;ac<vvw;ac++)
      {
        coord=dn+ac;
        r=(float)brightness(camera.pixels[coord])/256.0f;
        r= lerp(centerTime,edgeTime,max(0.0f,min(r,0.99f))) ;
        blendMap[coord]=(int) (((timeBufferSize-1)*blendLevels)*min(1.0f,r)) ;
      }
  }

  // checkery map
  if(functionType ==4)
  {
   int lastLevel = (timeBufferSize-1)*blendLevels;
    for( dn=0;dn<vvh;dn++)
      for( ac=0;ac<vvw;ac++)
      {
        coord=dn*vvw+ac;
        // some kind of function for fps
        // center should be slow, edge should be fast!

        r=(float)dn/(float)vvh;
        // blend amounts... hmmm!
         if((ac+dn)%2 == 0)
       // if(((int(ac/10)+int(dn/10)))%2 ==0)
        {
          blendMap[coord]=0;
        }
        else
        {
          blendMap[coord]=lastLevel; // * ((ac+dn)%16)/16;
        }
      }
  }

  //most likely map
  if(functionType ==5)
  {

    for( dn=0;dn<vvh*vvw;dn+=vvw)
      for( ac=0;ac<vvw;ac++)
      {
        coord=dn+ac;

        int sumColorB=0;
        for( tbuf = iBuffer;tbuf<iBuffer+likelySize;tbuf++)
        {
          tbuf0 =  tbuf % timeBufferSize ;
          sumColorB+=brightness(timeBuffer[tbuf0].pixels[coord]);
        };

        tbuf0 =  iBuffer % timeBufferSize ;
        likelyColorB=(int)brightness(timeBuffer[tbuf0].pixels[coord]);
        sumColorB/=likelySize;
        likelyIndex=-1;
        likelyVariance=0;
        for( tbuf =  iBuffer+1;tbuf<iBuffer+likelySize;tbuf++)
        {
          tbuf0 =  tbuf % timeBufferSize ;
          thatBrt = (int)brightness(timeBuffer[tbuf0].pixels[coord]);
          likelyVariance=max(thatBrt,likelyVariance);
          if( abs(thatBrt-sumColorB)<likelyColorB)
          {
            likelyColorB=thatBrt;
            likelyIndex=tbuf0;
          }
        }
        
  if(likelyVariance>1)
  {
        blendMap[coord]=(int) ((likelyIndex*blendLevels) ) ;
  }
  else
  {
    blendMap[coord]=-1;
  }
        //  if(r<0.10)
        //  println("r:"+r+" map:"+blendMap[coord]);
      }
  }

}


public void captureEvent(Capture camera)
{
  camera.read();
  timeBuffer[iBuffer].copy(camera, 0, 0, vvw,  vvh, 0, 0, vvw, vvh);

  iBuffer++;
  if(iBuffer>=timeBufferSize)
    iBuffer=0;
  newFrame = true;

}



public void draw()
{
  int c0,c1,clerp;
  int RR,GG,BB;
  int coord;
  int tbuf;
  int tbuf0;


  if(newFrame) {
    //   textFont(myFont); 
    noStroke();
    if(blendType==0)
    {

    }
    if((blendType==1)&&(functionType==0))
    {
      cw=mouseX;
      ch=mouseY;
      setupBlendMap();
    }
    if(functionType==5)
    {
      setupBlendMap();
    }

    for(int down=0;down<vvh;down++)
    {
      for(int across=0;across<vvw;across++)
      {
        coord = down*vvw+across;
        // neg map = substitute a standard color
        if(blendMap[coord]>=0)
        {
          tbuf = (blendMap[coord]>>blendLevelsL2)+iBuffer;
          tbuf0 =  tbuf % timeBufferSize ;

          //int tlerp = blendMap[coord]&blendLevelsMASK;
          // int tbuf1 = min(tbuf+1,timeBufferSize-1);
          //  c0=timeBuffer[tbuf1].pixels[coord];
          // c1=timeBuffer[tbuf0].pixels[coord];
          //  RR= (int)( (red(c0)*tlerp) + (red(c1)*(blendLevelsMASK-tlerp) )) >>blendLevelsL2;
          //   GG= (int)( (green(c0)*tlerp) + (green(c1)*(blendLevelsMASK-tlerp) )) >>blendLevelsL2;
          // BB= (int)( (blue(c0)*tlerp) + (blue(c1)*(blendLevelsMASK-tlerp) )) >>blendLevelsL2;
          //clerp=color(RR,GG,BB);
          //   set(across, down, clerp );

          set(across, down, timeBuffer[tbuf0].pixels[coord] );
        }
        else
        {
          set(across, down, 0xFFF08040 );
        }

      }
    }
    drawLabel();
    // for(int down=0;down<vvh;down++)
    //text(iBuffer, 10, 50); 
    // image(timeBuffer,0,0);
    if(saving==1)
    {
      saveFrame("delay/####.tif");
    }
    newFrame = false;
  }
}

public void keyPressed(){
  println("key:"+keyCode+ "  key:"+key);
  if (keyCode == 16)
  { // press SHIFT to restart saving
    saving = 1-saving;
  }
  if(key == '1')
  {
    functionType=0;
    setupBlendMap();
  }
  if(key == '2')
  {
    functionType=1;
    setupBlendMap();
    blendType=0;
  }
  if(key == '3')
  {
    functionType=2;
    setupBlendMap();
    blendType=0;
  }  

  if(key == '4')
  {
    functionType=3;
    setupBlendMap();
    blendType=0;
  }
  if(key  == '5')
  {
    functionType=4;
    setupBlendMap();
    blendType=0;
  }
  if(key  == '6')
  {
    functionType=5;
    setupBlendMap();
    blendType=0;
  }
  if(key == 'n')
  {
    blendType=0;
  }
  if(key == 'm')
  {
    blendType=1;
  }

}


public void drawLabel()
{
  String delayModeString,delayTypeString,delaySavingString;

  fill(240,240,240);
  rect(0,svh,svw,svh+20);
  fill(64,64,64);

  switch(functionType)
  {
  case 0: 
    delayModeString="Circular";
    break;
  case 1: 
    delayModeString="Left+Right";
    break;
  case 2: 
    delayModeString="Bottom to Top";
    break;
  case 3: 
    delayModeString="Image Brightness";
    break;
  case 4: 
    delayModeString="Checkers";
    break;
  case 5: 
    delayModeString="MostLikely";
    break;
  default:
    delayModeString="?";
  }

  switch(blendType)
  {
  case 0: 
    delayTypeString="center";
    break;
  case 1: 
    delayTypeString="mouse";
    break;
  default:
    delayTypeString="?";

  }

  delaySavingString="-";
  switch(saving)
  {
  case 0: 
    delaySavingString="-";
    break;
  case 1: 
    delaySavingString="Saving";
    break;
  default:
    delaySavingString="?";

  }



  text(delayModeString+" "+delayTypeString+" "+delaySavingString,15,svh+14);
}

  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "--full-screen", "--bgcolor=#666666", "--stop-color=#cccccc", "DelayVideo" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
